@@ -1,4 +1,5 @@
 import logging
+import threading
 import uuid
 
 from opentelemetry.sdk.trace.export import (
@@ -30,6 +31,8 @@ class LaminarSpanProcessor(SpanProcessor):
     __span_id_to_path: dict[int, list[str]] = {}
     __span_id_lists: dict[int, list[str]] = {}
     max_export_batch_size: int
+    _instance_lock: threading.RLock
+    _paths_lock: threading.RLock
 
     def __init__(
         self,
@@ -42,6 +45,8 @@ class LaminarSpanProcessor(SpanProcessor):
         disable_batch: bool = False,
         exporter: SpanExporter | None = None,
     ):
+        self._instance_lock = threading.RLock()
+        self._paths_lock = threading.RLock()
         self.logger = get_default_logger(__name__)
         self.max_export_batch_size = max_export_batch_size
         self.exporter = exporter or LaminarSpanExporter(
@@ -60,20 +65,25 @@ class LaminarSpanProcessor(SpanProcessor):
         )
 
     def on_start(self, span: Span, parent_context: Context | None = None):
-        parent_span_path = list(span.attributes.get(PARENT_SPAN_PATH, tuple())) or (
-            self.__span_id_to_path.get(span.parent.span_id) if span.parent else None
-        )
-        parent_span_ids_path = list(
-            span.attributes.get(PARENT_SPAN_IDS_PATH, tuple())
-        ) or (self.__span_id_lists.get(span.parent.span_id, []) if span.parent else [])
-        span_path = parent_span_path + [span.name] if parent_span_path else [span.name]
-        span_ids_path = parent_span_ids_path + [
-            str(uuid.UUID(int=span.get_span_context().span_id))
-        ]
-        span.set_attribute(SPAN_PATH, span_path)
-        span.set_attribute(SPAN_IDS_PATH, span_ids_path)
-        self.__span_id_to_path[span.get_span_context().span_id] = span_path
-        self.__span_id_lists[span.get_span_context().span_id] = span_ids_path
+        with self._paths_lock:
+            parent_span_path = list(span.attributes.get(PARENT_SPAN_PATH, tuple())) or (
+                self.__span_id_to_path.get(span.parent.span_id) if span.parent else None
+            )
+            parent_span_ids_path = list(
+                span.attributes.get(PARENT_SPAN_IDS_PATH, tuple())
+            ) or (
+                self.__span_id_lists.get(span.parent.span_id, []) if span.parent else []
+            )
+            span_path = (
+                parent_span_path + [span.name] if parent_span_path else [span.name]
+            )
+            span_ids_path = parent_span_ids_path + [
+                str(uuid.UUID(int=span.get_span_context().span_id))
+            ]
+            span.set_attribute(SPAN_PATH, span_path)
+            span.set_attribute(SPAN_IDS_PATH, span_ids_path)
+            self.__span_id_to_path[span.get_span_context().span_id] = span_path
+            self.__span_id_lists[span.get_span_context().span_id] = span_ids_path
 
         span.set_attribute(SPAN_INSTRUMENTATION_SOURCE, "python")
         span.set_attribute(SPAN_SDK_VERSION, __version__)
@@ -84,13 +94,16 @@ class LaminarSpanProcessor(SpanProcessor):
             for key, value in graph_context.items():
                 span.set_attribute(f"lmnr.association.properties.{key}", value)
 
-        self.instance.on_start(span, parent_context)
+        with self._instance_lock:
+            self.instance.on_start(span, parent_context)
 
     def on_end(self, span: Span):
-        self.instance.on_end(span)
+        with self._instance_lock:
+            self.instance.on_end(span)
 
     def force_flush(self, timeout_millis: int = 30000) -> bool:
-        return self.instance.force_flush(timeout_millis)
+        with self._instance_lock:
+            return self.instance.force_flush(timeout_millis)
 
     def force_reinit(self):
         if not isinstance(self.exporter, LaminarSpanExporter):
@@ -98,23 +111,32 @@ class LaminarSpanProcessor(SpanProcessor):
                 "LaminarSpanProcessor is not using LaminarSpanExporter, cannot force reinit"
             )
             return
-        self.instance.shutdown()
-        disable_batch = isinstance(self.instance, SimpleSpanProcessor)
-        del self.exporter.instance
-        del self.instance
 
+        # Reinitialize exporter (thread-safe, handles its own locking)
         self.exporter._init_instance()
-        self.instance = (
-            SimpleSpanProcessor(self.exporter)
-            if disable_batch
-            else BatchSpanProcessor(
-                self.exporter, max_export_batch_size=self.max_export_batch_size
+
+        with self._instance_lock:
+            old_instance = self.instance
+            disable_batch = isinstance(old_instance, SimpleSpanProcessor)
+
+            try:
+                old_instance.shutdown()
+            except Exception as e:
+                self.logger.debug(f"Error shutting down old processor instance: {e}")
+
+            self.instance = (
+                SimpleSpanProcessor(self.exporter)
+                if disable_batch
+                else BatchSpanProcessor(
+                    self.exporter, max_export_batch_size=self.max_export_batch_size
+                )
             )
-        )
 
     def shutdown(self):
-        self.instance.shutdown()
+        with self._instance_lock:
+            self.instance.shutdown()
 
     def clear(self):
-        self.__span_id_to_path = {}
-        self.__span_id_lists = {}
+        with self._paths_lock:
+            self.__span_id_to_path = {}
+            self.__span_id_lists = {}
